
#ifndef PLS_BACKOFF_H_
#define PLS_BACKOFF_H_

#include "pls/internal/base/system_details.h"
#include "pls/internal/helpers/profiler.h"
#include "pls/internal/base/thread.h"

#include <random>
#include <math.h>

namespace pls {
namespace internal {
namespace base {

class backoff {
  static constexpr unsigned long INITIAL_SPIN_ITERS = 2u << 2u;
  static constexpr unsigned long MAX_SPIN_ITERS = 2u << 6u;

  unsigned long current_ = INITIAL_SPIN_ITERS;
  std::minstd_rand random_;

  static void spin(unsigned long iterations) {
    for (volatile unsigned long i = 0; i < iterations; i++)
      system_details::relax_cpu(); // Spin
  }

 public:
  backoff() : current_{INITIAL_SPIN_ITERS}, random_{std::random_device{}()} {}

  void do_backoff() {
    PROFILE_LOCK("Backoff")
    spin(random_() % std::min(current_, MAX_SPIN_ITERS));

    current_ = current_ * 2;
    if (current_ > MAX_SPIN_ITERS) {
      current_ = MAX_SPIN_ITERS;
    }
  }

  void reset() {
    current_ = INITIAL_SPIN_ITERS;
  }
};

}
}
}

#endif //PLS_BACKOFF_H_
